#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <unistd.h>

#include "util.h"
#include "cable.h"
#include "target.h"
#include "jtag.h"

extern char *progname;
extern int verbose;

extern char *conf_getenv(char *name);

static void usage() {
	printf("cmd mode usage:\n");
	printf("   (fl|file_load) [filetype] file\n");
	printf("      filetype: bit jed mcs mem\n");
	printf("   (tl|target_load) memtype [size]\n");
	printf("      memtype: fpga_ram  fpga_flash spi_flash rtavr_isp\n");
	printf("               ram       flash      spi       isp\n");
	printf("   (d|dump) [addr size flags] [! outfile]\n");
	printf("      flags: 1 reverse bit (in byte)\n");
	printf("   (e|erase)   memtype\n");
	printf("   (w|program) memtype\n");
	printf("   (v|verify)  memtype\n");
	printf("   (t|terminal) [bitclock]\n");
	printf("   (q|quit)\n");
}

extern void jtag_test(CABLE cbl);
extern void spi_test(TARGET tgt, CABLE cbl);
extern void isp_test(TARGET tgt, CABLE cbl);
extern void rd1107_test(TARGET tgt, CABLE cbl);

extern int bit_load(FILE *fp, uint8_t *buf, int bufsiz);
extern int jed_load(FILE *fp, uint8_t *buf, int bufsiz);
extern int mcs_load(FILE *fp, uint8_t *buf, int bufsiz);
extern int mem_load(FILE *fp, uint8_t *buf, int bufsiz);

#define FILETYPE_NULL		0
#define FILETYPE_BIT		1
#define FILETYPE_JED		2
#define FILETYPE_MCS		3
#define FILETYPE_MEM		4

int cur_filetype = FILETYPE_NULL;

static int filetype(char *str) {
	int ft = FILETYPE_NULL;
	if (!strcmp(str, "bit")) {
		ft = FILETYPE_BIT;
	} else if (!strcmp(str, "jed")) {
		ft = FILETYPE_JED;
	} else if (!strcmp(str, "mcs")) {
		ft = FILETYPE_MCS;
	} else if (!strcmp(str, "mem")) {
		ft = FILETYPE_MEM;
	}
	// alt definition
	else if (!strcmp(str, "hex")) {
		ft = FILETYPE_MCS;
	}
	return ft;
}

static int file_load(int ft, FILE *fp, uint8_t *buf, int bufsiz) {
	int ret = -1;
	fflush(stdout);
	if (ft == FILETYPE_BIT) {
		ret = bit_load(fp, buf, bufsiz);
	} else if (ft == FILETYPE_JED) {
		ret = jed_load(fp, buf, bufsiz);
	} else if (ft == FILETYPE_MCS) {
		ret = mcs_load(fp, buf, bufsiz);
	} else if (ft == FILETYPE_MEM) {
		ret = mem_load(fp, buf, bufsiz);
	} else {
		ret = -1;
	}
	fflush(stderr);
	fflush(stdout);
	return ret;
}

extern int isp_read(TARGET tgt, CABLE cbl, uint8_t *buf, int len);
extern int isp_program(TARGET tgt, CABLE cbl, uint8_t *buf, int len);

#define TARGET_NULL		0
#define TARGET_FPGA_CONF	1
#define TARGET_FPGA_FLASH	2
#define TARGET_SPI_FLASH	3
#define TARGET_RTAVR_ISP	4

static uint32_t xtoi(char *str) {
	uint32_t v = 0;
	if (!strncasecmp(str, "0x", 2)) {
		sscanf(str+2, "%x", &v);
	} else {
		sscanf(str, "%d", &v);
	}
	return v;
}

int cur_memtype = TARGET_NULL;
static int memtype(char *str) {
	int target = TARGET_NULL;
	if (!strcmp(str, "fpga_ram")) {
		target = TARGET_FPGA_CONF;
	} else if (!strcmp(str, "fpga_flash")) {
		target = TARGET_FPGA_FLASH;
	} else if (!strcmp(str, "spi_flash")) {
		target = TARGET_SPI_FLASH;
	} else if (!strcmp(str, "rtavr_isp")) {
		target = TARGET_RTAVR_ISP;
	}
	// short definition
	else if (!strcmp(str, "ram")) {
		target = TARGET_FPGA_CONF;
	} else if (!strcmp(str, "flash")) {
		target = TARGET_FPGA_FLASH;
	} else if (!strcmp(str, "spi")) {
		target = TARGET_SPI_FLASH;
	} else if (!strcmp(str, "isp")) {
		target = TARGET_RTAVR_ISP;
	}
	return target;
}

static int target_load(int target, TARGET tgt, CABLE cbl, uint8_t *buf, int bufsiz) {
	if (target == TARGET_FPGA_CONF) {
		if (!tgt->read_conf) return -1;
		return (tgt->read_conf)(tgt, cbl, buf, bufsiz);
	} else if (target == TARGET_FPGA_FLASH) {
		if (!tgt->read_flash) return -1;
		return (tgt->read_flash)(tgt, cbl, buf, bufsiz);
	} else if (target == TARGET_SPI_FLASH) {
	} else if (target == TARGET_RTAVR_ISP) {
		return isp_read(tgt, cbl, buf, bufsiz);
	}
	return -1;
}

static int target_program(int target, TARGET tgt, CABLE cbl, uint8_t *buf, int bufsiz) {
	if (target == TARGET_FPGA_CONF) {
		if (!tgt->write_conf) return -1;
		return (tgt->write_conf)(tgt, cbl, buf, bufsiz);
	} else if (target == TARGET_FPGA_FLASH) {
		if (!tgt->write_flash) return -1;
		return (tgt->write_flash)(tgt, cbl, buf, bufsiz);
	} else if (target == TARGET_SPI_FLASH) {
	} else if (target == TARGET_RTAVR_ISP) {
		return isp_program(tgt, cbl, buf, bufsiz);
	}
	return -1;
}

static int target_erase(int target, TARGET tgt, CABLE cbl) {
	if (target == TARGET_FPGA_FLASH) {
		if (!tgt->erase_flash) return -1;
		return (tgt->erase_flash)(tgt, cbl);
	} else if (target == TARGET_SPI_FLASH) {
	} else if (target == TARGET_RTAVR_ISP) {
		//return isp_erase(tgt, cbl, buf, bufsiz);
	}
	return -1;
}

static void cmd_parse(char *line, int *argc, char *argv[]) {
	int i;
	char *p = line;
	(*argc) = 0;
	for (i=0; i<10; i++) {
		while ((*p == ' ') || (*p == '\t')) p++;
		if ((*p == 0) || (*p == '#')) break;
		*argv++ = p;
		*argc = i+1;
		while (*p && !((*p == ' ') || (*p == '\t'))) p++;
		if (*p) *p++ = 0;
	}
}

#define MAX_MEM_SIZE	(1024 * 1024 * 4)
uint8_t *mem_buf;
int mem_size = -1;
uint8_t *alt_buf;
int alt_size = -1;

int cmd_mode(TARGET tgt, CABLE cbl) {
	int ret = 0;
	int len;
	int ft;
	int target;
	FILE *fp = NULL;
	char *p = NULL;
	char line[256];
	int argc;
	char *argv[10];
	uint32_t d_addr = 0, d_size = 0;
	int d_flags = 0;

        mem_buf = malloc(MAX_MEM_SIZE * 2);
	if (!mem_buf) {
		fprintf(stderr, "malloc error!\n");
		exit(1);
	}
        memset(mem_buf, 0xff, MAX_MEM_SIZE * 2);
        alt_buf = mem_buf + MAX_MEM_SIZE;

	printf("cmd> ");
	fflush(stdout);
	while (fgets(line, 256, stdin)) {
		len = strlen(line);
		if (line[len-1] == '\n') len--;
		if (line[len-1] == '\r') len--;
		line[len] = 0;
		cmd_parse(line, &argc, argv);
		if (argc == 0) goto next;
		if (argc == 1) {
		    if (!strcmp(argv[0], "help") ||  !strcmp(argv[0], "?")) {
			usage();
		    } else if (!strcmp(argv[0], "quit") || !strcmp(argv[0], "q")) {
			break;
		    } else if (!strcmp(argv[0], "dump") || !strcmp(argv[0], "d")) {
			// dump [addr size flags] [! outfile]
			if (mem_size < 0) {
				printf("  memory is not valid\n");
				ret = mem_size;
				goto next;
			}
			printf("*** dump *** addr 0x%04x size 0x%04x(%d) flags 0x%x(%s)\n"
					,d_addr,d_size, d_size, d_flags 
					,(d_flags&1)? "REV":"NML");
			dump(mem_buf, d_addr, d_size, d_flags, stdout);
			if (d_addr + d_size <= mem_size)
				d_addr += d_size;
		    } else if (!strcmp(argv[0], "terminal")) {
			// terminal [bitclock]
		    } else if (!strcmp(argv[0], "spi_test")) {
			spi_test(tgt, cbl);
		    } else if (!strcmp(argv[0], "isp_test")) {
			isp_test(tgt, cbl);
		    } else if (!strcmp(argv[0], "pwm_test")) {
			rd1107_test(tgt, cbl);
		    } else if (!strcmp(argv[0], "jtag_test")) {
			jtag_test(cbl);
		    } else {
			goto invalid_cmd;
		    }
		} else if (argc == 2) {
		    target = memtype(argv[1]);
		    if (!strcmp(argv[0], "program") || !strcmp(argv[0], "w")) {
			// program memtype
			if (mem_size < 0) {
				ret = mem_size;
				printf("  memory is not valid\n");
				goto next;
			}
			if (!target) {
				ret = -1;
				printf("  unkown memtype %s\n",argv[1]);
				goto next;
			}
			ret = target_program(target, tgt, cbl, mem_buf, mem_size);
			if (ret < 0) {
				printf("  program error\n");
				goto next;
			}
			printf("  program success\n");
		    } else if (!strcmp(argv[0], "erase") || !strcmp(argv[0], "e")) {
			// erase memtype
			if (!target) {
				ret = -1;
				printf("  unkown memtype %s\n",argv[1]);
				goto next;
			}
			ret = target_erase(target, tgt, cbl);
			if (ret < 0) {
				printf("  erase error\n");
				goto next;
			}
			printf("  erase success\n");
		    } else if (!strcmp(argv[0], "verify") || !strcmp(argv[0], "v")) {
			// verify memtype
			int i;
			if (mem_size < 0) {
				ret = mem_size;
				printf("  memory is not valid\n");
				goto next;
			}
			if (!target) {
				ret = -1;
				printf("  unkown memtype %s\n",argv[1]);
				goto next;
			}
			alt_size = target_load(target, tgt, cbl, alt_buf, mem_size);
			if (alt_size < 0) {
				ret = alt_size;
				printf("  target load error\n");
				goto next;
			}
			for (i=0; i< alt_size; i++) {
				if (mem_buf[i] != alt_buf[i]) {
					printf("  verify error at %d 0x%02x != 0x%02x\n", i, mem_buf[i], alt_buf[i]);
					ret = -1;
					goto next;
				}
			}
			if (mem_size != alt_size) {
				printf("  verify error at %d target size is short\n", i);
				ret = -1;
				goto next;
			}
			printf("  verify suscess\n");
		    } else if (!strcmp(argv[0], "file_load")
		              || !strcmp(argv[0], "fl")) {
			// file_load [filetype] file
			p = strrchr(argv[1],'.');
			if (!p) {
				printf("  can't guess file type\n");
				goto next;
			}
			ft = filetype(p+1);
			if (!ft) {
				printf("  unkown file type %s\n",p+1);
				goto next;
			}
			fp = fopen(argv[1],"rb");
			if (!fp) {
				printf("  can't open file\n");
				goto next;
			}
			mem_size = file_load(ft, fp, mem_buf, MAX_MEM_SIZE);
			fclose(fp);
			if (mem_size < 0) {
				printf("  memory is not valid\n");
				goto next;
			}
			printf("  memory loaded size = %d\n", mem_size);
			d_addr = 0; d_size = mem_size;
		    } else if (!strcmp(argv[0], "target_load")
		              || !strcmp(argv[0], "tl")) {
			// target_load memtype [size]
			int size = MAX_MEM_SIZE;
		        target = memtype(argv[1]);
			if (!target) {
				printf("  unkown memtype %s\n",argv[1]);
				goto next;
			}
			if (verbose)
				fprintf(stderr, "  size = 0x%x(%d)\n", size, size);
			mem_size = target_load(target, tgt, cbl, mem_buf, size);
			if (mem_size < 0) {
				printf("  memory is not valid\n");
				goto next;
			}
			printf("  memory loaded size = %d\n", mem_size);
			d_addr = 0;
			d_size = mem_size;
		    } else {
			goto invalid_cmd;
		    }
		} else if (argc == 3) {
		    if (!strcmp(argv[0], "file_load")
		              || !strcmp(argv[0], "fl")) {
			// file_load [filetype] file
			ft = filetype(argv[1]);
			if (!ft) {
				printf("  unkown file type %s\n",p+1);
				goto next;
			}
			fp = fopen(argv[2],"rb");
			if (!fp) {
				printf("  can't open file\n");
				goto next;
			}
			mem_size = file_load(ft, fp, mem_buf, MAX_MEM_SIZE);
			if (mem_size < 0) {
				printf("  memory is not valid\n");
			}
			fclose(fp);
			printf("  memory loaded size = %d\n", mem_size);
			d_addr = 0; d_size = mem_size;
		    } else if (!strcmp(argv[0], "target_load")
		              || !strcmp(argv[0], "tl")) {
			// target_load memtype [size]
		        int size;
		        target = memtype(argv[1]);
		        size = xtoi(argv[2]);
			if (verbose)
				fprintf(stderr, "  size = 0x%x(%d)\n", size, size);
			mem_size = target_load(target, tgt, cbl, mem_buf, size);
			if (mem_size < 0) {
				printf("  memory is not valid\n");
				goto next;
			}
			printf("  memory loaded size = %d\n", mem_size);
			d_addr = 0;
			d_size = mem_size;
		    } else if (!strcmp(argv[0], "dump") || !strcmp(argv[0], "d")) {
		        FILE *out_fp;
			// dump [addr size flags] [! outfile]
			// dump ! outfile
			if (mem_size <= 0) {
				printf("  memory is not valid\n");
				goto next;
			}
			if (strcmp(argv[1], "!")) {
				goto invalid_cmd;
			}
			out_fp = fopen(argv[2],"a");
			if (!out_fp) {
				printf("  can't open output file\n");
				goto next;
			}
			fflush(stdout);
			fflush(stderr);
			fprintf(out_fp, "*** dump *** addr 0x%04x size 0x%04x(%d) flags 0x%x(%s)\n"
					,d_addr,d_size, d_size, d_flags 
					,(d_flags&1)? "REV":"NML");
			dump(mem_buf, d_addr, d_size, d_flags, out_fp);
			fclose(out_fp);
			if (d_addr + d_size <= mem_size)
				d_addr += d_size;
		    } else {
			goto invalid_cmd;
		    }
		} else if (argc == 4) {
		    uint32_t addr,size,flags;
		    if (!strcmp(argv[0], "dump") || !strcmp(argv[0], "d")) {
			// dump [addr size flags] [! outfile]
			addr = xtoi(argv[1]);
			size = xtoi(argv[2]);
			flags = xtoi(argv[3]);
			if (mem_size <= 0) {
				printf("  memory is not valid\n");
				goto next;
			}
			if ((size <= 0) || (addr >= MAX_MEM_SIZE)) {
				printf("  addr or size is invalid value\n");
				goto next;
			}
			fflush(stdout);
			fflush(stderr);
			printf("*** dump *** addr 0x%04x size 0x%04x(%d) flags 0x%x(%s)\n"
					,addr,size, size, flags 
					,(flags&1)? "REV":"NML");
			dump(mem_buf, addr, size, flags, stdout);
			if (addr + size <= mem_size) {
				d_flags = flags;
				d_size = size;
				d_addr = addr + size;
			}
		    } else {
			goto invalid_cmd;
		    }
		} else if (argc == 6) {
		    FILE *out_fp;
		    uint32_t addr,size,flags;
		    if (!strcmp(argv[0], "dump") || !strcmp(argv[0], "d")) {
			// dump [addr size flags] [! outfile]
			addr = xtoi(argv[1]);
			size = xtoi(argv[2]);
			flags = xtoi(argv[3]);
			if (strcmp(argv[4], "!")) {
				goto invalid_cmd;
			}
			out_fp = fopen(argv[5],"a");
			if (!out_fp) {
				printf("  can't open output file\n");
				goto next;
			}
			if (mem_size <= 0) {
				printf("  memory is not valid\n");
				goto next;
			}
			if ((size <= 0) || (addr >= MAX_MEM_SIZE)) {
				printf("  addr or size is invalid value\n");
				goto next;
			}
			fflush(stdout);
			fflush(stderr);
			fprintf(out_fp, "*** dump *** addr 0x%04x size 0x%04x(%d) flags 0x%x(%s)\n"
					,addr,size, size, flags 
					,(flags&1)? "REV":"NML");
			dump(mem_buf, addr, size, flags, out_fp);
			fclose(out_fp);
			if (addr + size <= mem_size) {
				d_flags = flags;
				d_size = size;
				d_addr = addr + size;
			}
		    } else {
			goto invalid_cmd;
		    }
		} else {
			int i;
invalid_cmd:

			printf("  invaled cmd");
			for (i=0; i< argc; i++) 
				printf(" '%s'", argv[i]);
			printf("\n");
		}
next:
		fflush(stdout);
		fflush(stderr);
		printf("cmd> ");
		fflush(stdout);
	}
	return 0;
}
